@@ -2,6 +2,7 @@
 # -*- coding: utf-8 -*-
 import sys
 import re
+import copy
 
 from dup.log import log
 
@@ -399,7 +400,7 @@ function showOutput(id, name) {
 
 
 class _TestResult(unittest.TestResult):
-    def __init__(self, verbosity=1):
+    def __init__(self, verbosity=1, retry=0, save_last_retry=True):
         super(_TestResult, self).__init__(verbosity)
         self.outputBuffer = io.BytesIO()
         self.raw_stdout = None
@@ -408,7 +409,11 @@ class _TestResult(unittest.TestResult):
         self.failure_count = 0
         self.skip_count = 0
         self.error_count = 0
-        self.verbosity = verbosity
+        self.status = 0
+        self.verbosity = verbosity,
+        self.retry = retry
+        self.trys = 0
+        self.save_last_retry = save_last_retry
         self.result = []
         self._case_start_time = 0
         self._case_run_time = 0
@@ -433,16 +438,41 @@ class _TestResult(unittest.TestResult):
         return result
 
     def stopTest(self, test):
+        if self.retry:
+            if self.status == 1:
+                self.trys += 1
+                if self.trys <= self.retry:
+                    if self.save_last_retry:
+                        t = self.result.pop(-1)
+                        if t[0] == 1:
+                            self.failure_count -= 1
+                        else:
+                            self.error_count -= 1
+                    test = copy.copy(test)
+                    sys.stderr.write("Retesting... ")
+                    sys.stderr.write(str(test))
+                    sys.stderr.write('..%d \n' % self.trys)
+                    doc = test._testMethodDoc or ''
+                    if doc.find('_retry') != -1:
+                        doc = doc[:doc.find('_retry')]
+                    desc = "%s_retry:%d" % (doc, self.trys)
+                    test._testMethodDoc = desc
+                    test(self)
+                else:
+                    self.status = 0
+                    self.trys = 0
         self.complete_output()
 
     def addSuccess(self, test):
         self.success_count += 1
+        self.status = 0
         super(_TestResult, self).addSuccess(test)
         output = self.complete_output()
         self.result.append((0, test, output, '', self._case_run_time))
 
     def addError(self, test, err):
         self.error_count += 1
+        self.status = 1
         super(_TestResult, self).addError(test, err)
         _, _exc_str = self.errors[-1]
         output = self.complete_output()
@@ -458,6 +488,7 @@ class _TestResult(unittest.TestResult):
 
     def addFailure(self, test, err):
         self.failure_count += 1
+        self.status = 1
         super(_TestResult, self).addFailure(test, err)
         _, _exc_str = self.failures[-1]
         output = self.complete_output()
@@ -468,38 +499,44 @@ class _TestResult(unittest.TestResult):
 
 
 class BSTestRunner(Template_mixin):
-    def __init__(self, report_title, report_dir, verbosity=1, description=""):
+    def __init__(self, report_title, report_dir, verbosity=1, description="", retry=0, save_last_try=True):
         self.report_dir = report_dir
         self.verbosity = verbosity
         self.title = report_title
         self.description = description
+        self.retry = retry
+        self.save_last_retry = save_last_try
         self.start_time = datetime.datetime.now()
         self.stop_time = None
 
     def run(self, test):
         log.info("开始进行测试")
-        result = _TestResult(self.verbosity)
+        result = _TestResult(self.verbosity, retry=self.retry, save_last_retry=self.save_last_retry)
         test(result)
         self.stop_time = datetime.datetime.now()
         self.generate_report(result)
         log.info('Time Elapsed: {}'.format(self.stop_time - self.start_time))
 
-        if Setting.create_ztest_style_report:
-            file = os.path.join(self.report_dir, r"{}-ztest.html".format(
-                self.start_time.strftime("%Y-%m-%d-%H-%M-%S")))
+        try:
+            if Setting.create_ztest_style_report:
+                file = os.path.join(self.report_dir, r"{}-test.html".format(self.title))
 
-            shutil.copy2(
-                os.path.join(os.path.dirname(__file__),
-                             "templates/template.html"),
-                file)
+                shutil.copy2(
+                    os.path.join(os.path.dirname(__file__),
+                                 "templates/template.html"),
+                    file)
 
-            with open(file, "r+") as f:
-                content = f.read().replace(r"${resultData}",
-                                           json.dumps(result_data,
-                                                      ensure_ascii=False,
-                                                      indent=4))
-                f.seek(0)
-                f.write(content)
+                with open(file, "r+") as f:
+                    content = f.read().replace(r"${resultData}",
+                                               json.dumps(result_data,
+                                                          ensure_ascii=False,
+                                                          indent=4))
+                    f.seek(0)
+                    f.write(content)
+        except Exception as e:
+            log.info("dup html Exception:{0}".format(e))
+        finally:
+            return result_data
 
     @staticmethod
     def sort_result(case_results):
@@ -653,6 +690,8 @@ class BSTestRunner(Template_mixin):
         else:
             case_tr_id = ""
         name = t.id().split('.')[-1]
+        method_name = [v for i, v in enumerate(name.split('_')) if i != 1]
+        name = '_'.join(method_name)
         doc = t.shortDescription() or ""
         desc = doc and ('%s: %s' % (name, doc)) or name
         tmpl_pass = has_output and self.REPORT_TEST_WITH_OUTPUT_TMPL_PASS or self.REPORT_TEST_NO_OUTPUT_TMPL
@@ -667,7 +706,7 @@ class BSTestRunner(Template_mixin):
         case_data['className'] = current_class_name
         case_data['methodName'] = name
         case_data['person'] = person
-        case_data['spendTime'] = "{:.2}S".format(run_time)
+        case_data['spendTime'] = run_time
         case_data['description'] = doc
         case_data['log'] = o + e
         if self.STATUS[n] == "Pass":
